from __future__ import annotations

from abc import ABC, abstractmethod
from fractions import Fraction
from functools import reduce
from hashlib import sha256
from itertools import zip_longest
from operator import mul
from typing import (
    TYPE_CHECKING,
    Any,
    ByteString,
    Dict,
    Generator,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    get_type_hints,
    overload,
)

from asn1crypto.core import Asn1Value, Integer, OctetString, SequenceOf

from . import asn1 as _asn1
from .groups import ImageGroup, ImageValue, PreGroup, PreGroupValue

if TYPE_CHECKING:  # pragma: no cover
    lazy = property
else:
    from lazy import lazy


def zip_strict(*args: Iterable[Any]) -> Iterator[Any]:
    """
    Like zip, but raise an exception unless all iterables have the same size
    """

    sentinel = object()
    for things in zip_longest(*args, fillvalue=sentinel):
        if any(thing is sentinel for thing in things):
            raise ValueError("Not all iters finished at the same time")
        yield things


_T0 = TypeVar("_T0")


def prod(items: Iterable[_T0], initializer: Optional[_T0] = None) -> _T0:
    """
    Product function, like sum() but with multiplication.

    Args:
        items: items to multiply
        initializer: Optional initializer

    Returns:
        product
    """

    if initializer is None:
        return reduce(mul, items)
    else:
        return reduce(mul, items, initializer)


_T1 = TypeVar("_T1", bound="Asn1Object")


class Asn1Object(ABC):
    pvss: Pvss
    asn1: Asn1Value

    def __init__(self, pvss: Pvss, asn1: Asn1Value) -> None:
        self.pvss = pvss
        self.asn1 = asn1
        self._validate()

    @classmethod
    def from_der(cls: Type[_T1], pvss: Pvss, data: ByteString) -> _T1:
        return cls(pvss, get_type_hints(cls)["asn1"].load(data))

    @property
    def der(self) -> bytes:
        return self.asn1.dump()

    @property
    def params(self) -> SystemParameters:
        return self.pvss.params

    @abstractmethod
    def _validate(self) -> None:
        """
        """

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return False

        return self.der == other.der


_T2 = TypeVar("_T2", bound="SystemParameters")


class SystemParameters(Asn1Object):
    ALGO: str
    asn1: _asn1.SystemParameters

    def __new__(cls, pvss: Pvss, asn1: _asn1.SystemParameters) -> SystemParameters:
        algo = asn1["algorithm"].native
        impl: Type[SystemParameters]
        if algo == "qr_mod_p":
            from . import qr

            impl = qr.QrParameters
        elif algo == "ristretto_255":
            from . import ristretto_255

            impl = ristretto_255.Ristretto255Parameters
        else:
            raise ValueError(f"Algorithm {algo} not implemented")
        return cast(SystemParameters, super().__new__(impl))

    @classmethod
    def create(cls: Type[_T2], pvss: Pvss, params: Any) -> _T2:
        return cls(pvss, _asn1.SystemParameters({"algorithm": cls.ALGO, "parameters": params}))

    @property
    @abstractmethod
    def pre_group(self) -> PreGroup:
        """
        """

    @lazy
    def g(self) -> Tuple[ImageValue, ImageValue]:
        return self._make_gen("g_0"), self._make_gen("g_1")

    @lazy
    def G(self) -> Tuple[ImageValue, ImageValue]:
        return self._make_gen("G_0"), self._make_gen("G_1")

    def _validate(self) -> None:
        # Load groups to validate them
        self.img_group
        self.pre_group

    @property
    @abstractmethod
    def img_group(self) -> ImageGroup:
        """
        """

    @abstractmethod
    def _make_gen(self, seed: str) -> ImageValue:
        """
        """


class PrivateKey(Asn1Object):
    asn1: _asn1.PrivateKey

    def _validate(self) -> None:
        """
        """
        self.priv

    @classmethod
    def create(cls, pvss: Pvss, priv: PreGroupValue) -> PrivateKey:
        return cls(pvss, _asn1.PrivateKey({"priv": priv.asn1}))

    @classmethod
    def create_random(cls, pvss: Pvss) -> PrivateKey:
        return cls.create(pvss, pvss.params.pre_group.rand_nonzero)

    @lazy
    def priv(self) -> PreGroupValue:
        return self.pvss.params.pre_group(self.asn1["priv"])

    def pub(self, name: str) -> PublicKey:
        return PublicKey.create(
            self.pvss, name, (self.params.G[0] ** self.priv, self.params.G[1] ** self.priv)
        )


class PublicKey(Asn1Object):
    asn1: _asn1.PublicKey

    def _validate(self) -> None:
        """
        """
        self.pub
        if not self.name:
            raise ValueError()

    @classmethod
    def create(cls, pvss: Pvss, name: str, pub: Tuple[ImageValue, ImageValue]) -> PublicKey:
        return cls(
            pvss, _asn1.PublicKey({"name": str(name), "pub0": pub[0].asn1, "pub1": pub[1].asn1})
        )

    @lazy
    def name(self) -> str:
        return str(self.asn1["name"])

    @lazy
    def pub(self) -> Tuple[ImageValue, ImageValue]:
        return (
            self.params.img_group(self.asn1["pub0"]),
            self.params.img_group(self.asn1["pub1"]),
        )


class Secret(Asn1Object):
    asn1: _asn1.Secret

    @classmethod
    def create(cls, pvss: Pvss, secret: ImageValue) -> Secret:
        return cls(pvss, _asn1.Secret({"secret": secret.asn1}))

    @lazy
    def secret(self) -> ImageValue:
        return self.params.img_group(self.asn1["secret"])

    def _validate(self) -> None:
        """
        """
        self.secret

    @classmethod
    def reconstruct(cls, pvss: Pvss, private_key: PrivateKey) -> Secret:
        """
        Decrypt the re-encrypted shares with the private key and reconstruct the secret

        Args:
            pvss: Pvss object with public values
            der_private_key: Receiver's Private key

        Returns:
            Secret
        """

        if len(pvss.reencrypted_shares) < len(pvss.shares.coefficients):
            raise Exception(
                f"Need at least {len(pvss.shares.coefficients)} shares, only got {len(pvss.reencrypted_shares)}"
            )

        shares = {
            reenc_share.idx: reenc_share.elg_b * ((reenc_share.elg_a ** private_key.priv) ** -1)
            for reenc_share in pvss.reencrypted_shares
        }

        secret = prod(
            share0
            ** prod(
                (Fraction(idx1, idx1 - idx0) for idx1 in shares.keys() if idx0 != idx1),
                Fraction(1),
            )
            for idx0, share0 in shares.items()
        )

        return cls.create(pvss, secret)


class Share(Asn1Object):
    asn1: _asn1.Share

    @classmethod
    def create(
        cls,
        pvss: Pvss,
        pub_name: str,
        share: ImageValue,
        resp: Tuple[PreGroupValue, PreGroupValue],
    ) -> Share:
        return cls(
            pvss,
            _asn1.Share(
                {
                    "pub": pub_name,
                    "share": share.asn1,
                    "response_f0": resp[0].asn1,
                    "response_f1": resp[1].asn1,
                }
            ),
        )

    @lazy
    def pub_name(self) -> str:
        return cast(str, self.asn1["pub"].native)

    @lazy
    def pub(self) -> PublicKey:
        """
        """

        return self.pvss.user_public_keys[self.pub_name]

    @lazy
    def share(self) -> ImageValue:
        return self.params.img_group(self.asn1["share"])

    @lazy
    def resp(self) -> Tuple[PreGroupValue, PreGroupValue]:
        return (
            self.params.pre_group(self.asn1["response_f0"]),
            self.params.pre_group(self.asn1["response_f1"]),
        )

    def _validate(self) -> None:
        """
        """
        # XXX check if pubname in pvss
        self.share
        self.resp


class SharedSecret(Asn1Object):
    asn1: _asn1.SharedSecret

    @classmethod
    def create(
        cls,
        pvss: Pvss,
        shares: Iterable[Share],
        coeffs: Iterable[ImageValue],
        challenge: ByteString,
    ) -> SharedSecret:
        return cls(
            pvss,
            _asn1.SharedSecret(
                {
                    "shares": [share.asn1 for share in shares],
                    "coefficients": [coeff.asn1 for coeff in coeffs],
                    "challenge": bytes(challenge),
                }
            ),
        )

    @lazy
    def shares(self) -> List[Share]:
        return [Share(self.pvss, share) for share in cast(SequenceOf, self.asn1["shares"])]

    @lazy
    def coefficients(self) -> List[ImageValue]:
        return [
            self.params.img_group(coeff)
            for coeff in cast(SequenceOf, self.asn1["coefficients"])
        ]

    @lazy
    def digest(self) -> bytes:
        return bytes(cast(OctetString, self.asn1["challenge"]))

    @lazy
    def challenge(self) -> int:
        return int.from_bytes(self.digest, "big")

    def _validate(self) -> None:
        X = [
            prod(c ** (i ** j) for j, c in enumerate(self.coefficients))
            for i in range(1, len(self.shares) + 1)
        ]

        min_c = -self.challenge
        r = [
            (
                (self.params.g[0] ** share.resp[0])
                * (self.params.g[1] ** share.resp[1])
                * (xi ** min_c),
                (share.pub.pub[0] ** share.resp[0])
                * (share.pub.pub[1] ** share.resp[1])
                * (share.share ** min_c),
            )
            for xi, share in zip(X, self.shares)
        ]

        challenge = SharesChallenge.create(
            pvss=self.pvss,
            pubs=[share.pub for share in self.shares],
            coeffs=self.coefficients,
            commitments=X,
            shares=[share.share for share in self.shares],
            randoms=r,
        )

        if challenge.digest != self.digest:
            raise ValueError("Verification failed: could not compute same challenge")

    @classmethod
    def create_shared_secret(
        cls, pvss: Pvss, qualified_size: int
    ) -> Tuple[Secret, SharedSecret]:
        """
        Create a secret, split it and compute the encrypted shares.

        Args:
            pvss: Pvss object with public values
            qualified_size: Number of shares required to reconstruct the secret

        Returns:
            Random secret and the encrypted shares
        """

        # user public keys
        pub = pvss.user_public_keys.values()

        # polynomials, chosen from Z_q
        alpha = (
            Poly(
                (pvss.params.pre_group.rand for __ in range(qualified_size)),
                pvss.params.pre_group(0),
            ),
            Poly(
                (pvss.params.pre_group.rand for __ in range(qualified_size)),
                pvss.params.pre_group(0),
            ),
        )

        # secret to be split
        S = Secret.create(
            pvss, (pvss.params.G[0] ** alpha[0](0)) * (pvss.params.G[1] ** alpha[1](0))
        )

        # commitments for coeffs
        C = [
            (pvss.params.g[0] ** coeff[0]) * (pvss.params.g[1] ** coeff[1])
            for coeff in zip(alpha[0], alpha[1])
        ]

        # encrypted shares
        Y = [
            (pi.pub[0] ** alpha[0](i)) * (pi.pub[1] ** alpha[1](i))
            for i, pi in enumerate(pub, 1)
        ]

        # X_i computed by prover
        X = [
            (pvss.params.g[0] ** alpha[0](i)) * (pvss.params.g[1] ** alpha[1](i))
            for i in range(1, len(pub) + 1)
        ]

        # rand [0,q)
        k = [(pvss.params.pre_group.rand, pvss.params.pre_group.rand) for __ in range(len(pub))]

        # random commitments
        r = [
            (
                (pvss.params.g[0] ** ki[0]) * (pvss.params.g[1] ** ki[1]),
                (pi.pub[0] ** ki[0]) * (pi.pub[1] ** ki[1]),
            )
            for pi, ki in zip(pub, k)
        ]

        # challenge is computed by hash function
        challenge = SharesChallenge.create(pvss, pub, C, X, Y, r)
        c = challenge.challenge

        # response
        s = [(ki[0] + alpha[0](i) * c, ki[1] + alpha[1](i) * c) for i, ki in enumerate(k, 1)]

        shared_secret = SharedSecret.create(
            pvss=pvss,
            shares=[
                Share.create(pvss=pvss, pub_name=pub_key.name, share=share, resp=resp)
                for pub_key, share, resp in zip(pub, Y, s)
            ],
            coeffs=C,
            challenge=challenge.digest,
        )

        return S, shared_secret


class ReencryptedShare(Asn1Object):
    asn1: _asn1.ReencryptedShare

    @classmethod
    def create(
        cls,
        pvss: Pvss,
        idx: int,
        elg_a: ImageValue,
        elg_b: ImageValue,
        response_priv: PreGroupValue,
        response_v: Tuple[PreGroupValue, PreGroupValue],
        response_w: Tuple[PreGroupValue, PreGroupValue],
        challenge: ByteString,
    ) -> ReencryptedShare:
        return cls(
            pvss,
            _asn1.ReencryptedShare(
                {
                    "idx": idx,
                    "elg_a": elg_a.asn1,
                    "elg_b": elg_b.asn1,
                    "response_priv": response_priv.asn1,
                    "response_v0": response_v[0].asn1,
                    "response_v1": response_v[1].asn1,
                    "response_w0": response_w[0].asn1,
                    "response_w1": response_w[1].asn1,
                    "challenge": bytes(challenge),
                }
            ),
        )

    @lazy
    def idx(self) -> int:
        return int(cast(Integer, self.asn1["idx"]))

    @lazy
    def elg_a(self) -> ImageValue:
        return self.params.img_group(self.asn1["elg_a"])

    @lazy
    def elg_b(self) -> ImageValue:
        return self.params.img_group(self.asn1["elg_b"])

    @lazy
    def response_priv(self) -> PreGroupValue:
        return self.params.pre_group(self.asn1["response_priv"])

    @lazy
    def response_v(self) -> Tuple[PreGroupValue, PreGroupValue]:
        return (
            self.params.pre_group(self.asn1["response_v0"]),
            self.params.pre_group(self.asn1["response_v1"]),
        )

    @lazy
    def response_w(self) -> Tuple[PreGroupValue, PreGroupValue]:
        return (
            self.params.pre_group(self.asn1["response_w0"]),
            self.params.pre_group(self.asn1["response_w1"]),
        )

    @lazy
    def digest(self) -> bytes:
        return bytes(cast(OctetString, self.asn1["challenge"]))

    @lazy
    def challenge(self) -> int:
        return int.from_bytes(self.digest, "big")

    @property
    def share(self) -> Share:
        return self.pvss.shares.shares[self.idx - 1]

    def _validate(self) -> None:
        """`
        """

        minus_c = -self.challenge
        pub = self.share.pub

        rand_share = (
            (self.elg_b ** self.response_priv)
            * (self.pvss.receiver_public_key.pub[0] ** self.response_v[0])
            * (self.pvss.receiver_public_key.pub[1] ** self.response_v[1])
        ) * (self.pvss.shares.shares[self.idx - 1].share ** minus_c)

        rand_pub = ((self.params.G[0] * self.params.G[1]) ** self.response_priv) * (
            (pub.pub[0] * pub.pub[1]) ** minus_c
        )
        rand_elg_a = (
            (self.params.G[0] ** self.response_w[0]) * (self.params.G[1] ** self.response_w[1])
        ) * (self.elg_a ** minus_c)
        rand_id = (
            (self.elg_a ** self.response_priv)
            * (self.params.G[0] ** self.response_v[0])
            * (self.params.G[1] ** self.response_v[1])
        )

        challenge = ReencryptedChallenge.create(
            self.pvss, rand_pub, rand_share, rand_elg_a, rand_id
        )

        if challenge.digest != self.digest:
            raise ValueError("Verification failed: could not compute same challenge")

    @classmethod
    def reencrypt(cls, pvss: Pvss, private_key: PrivateKey) -> ReencryptedShare:
        """
        Decrypt a share of the encrypted secret with the private_key and
        re-encrypt it with another public key

        Args:
            pvss: Pvss object with public values
            private_key: A user's private key

        Returns:
            Re-encrypted share
        """

        # Locate our share
        for idx, enc_share in enumerate(pvss.shares.shares, 1):
            if enc_share.pub == private_key.pub(enc_share.pub_name):
                break
        else:
            raise ValueError("No matching public key found")

        # decrypt our share
        share = enc_share.share ** private_key.priv.inv

        # Reencrypt share with Elgamal encryption using the receiver's public key
        w = [pvss.params.pre_group.rand, pvss.params.pre_group.rand]
        elg_a = (pvss.params.G[0] ** w[0]) * (pvss.params.G[1] ** w[1])
        elg_b = (
            share
            * (pvss.receiver_public_key.pub[0] ** w[0])
            * (pvss.receiver_public_key.pub[1] ** w[1])
        )

        v = [-w[0] * private_key.priv, -w[1] * private_key.priv]
        kv = (pvss.params.pre_group.rand, pvss.params.pre_group.rand)
        kw = (pvss.params.pre_group.rand, pvss.params.pre_group.rand)
        kpi = pvss.params.pre_group.rand

        rand_share = (
            (elg_b ** kpi)
            * (pvss.receiver_public_key.pub[0] ** kv[0])
            * (pvss.receiver_public_key.pub[1] ** kv[1])
        )
        rand_pub = (pvss.params.G[0] * pvss.params.G[1]) ** kpi
        rand_elg_a = (pvss.params.G[0] ** kw[0]) * (pvss.params.G[1] ** kw[1])
        rand_id = (elg_a ** kpi) * (pvss.params.G[0] ** kv[0]) * (pvss.params.G[1] ** kv[1])

        challenge = ReencryptedChallenge.create(pvss, rand_pub, rand_share, rand_elg_a, rand_id)

        c = challenge.challenge
        resp_priv = kpi + private_key.priv * c
        resp_v = (kv[0] + v[0] * c, kv[1] + v[1] * c)
        resp_w = (kw[0] + w[0] * c, kw[1] + w[1] * c)

        return ReencryptedShare.create(
            pvss, idx, elg_a, elg_b, resp_priv, resp_v, resp_w, challenge.digest
        )


class Challenge(Asn1Object):
    def _validate(self) -> None:
        """
        Noop because this is never loaded
        """
        # XXX is still loaded in constructor. What to do?!

    @lazy
    def digest(self) -> bytes:
        """
        Compute the sha256 digest over the DER encoding
        """
        return sha256(self.der).digest()

    @lazy
    def challenge(self) -> int:
        """
        Convert the digest into an integer so it can be used in the algorithms
        """
        return int.from_bytes(self.digest, "big")


class SharesChallenge(Challenge):
    asn1: _asn1.SharesChallenge

    @classmethod
    def create(
        cls,
        pvss: Pvss,
        pubs: Iterable[PublicKey],
        coeffs: Iterable[ImageValue],
        commitments: Iterable[ImageValue],
        shares: Iterable[ImageValue],
        randoms: Iterable[Tuple[ImageValue, ImageValue]],
    ) -> SharesChallenge:
        """
        """
        return cls(
            pvss,
            _asn1.SharesChallenge(
                {
                    "parameters": pvss.params.asn1,
                    "coefficients": [coeff.asn1 for coeff in coeffs],
                    "users": [
                        {
                            "pub": pub.asn1,
                            "commitment": commitment.asn1,
                            "random_commitment": rand_commit.asn1,
                            "share": share.asn1,
                            "random_share": rand_share.asn1,
                        }
                        for pub, commitment, share, (rand_commit, rand_share) in zip_strict(
                            pubs, commitments, shares, randoms
                        )
                    ],
                }
            ),
        )


class ReencryptedChallenge(Challenge):
    asn1: _asn1.ReencryptedChallenge

    @classmethod
    def create(
        cls,
        pvss: Pvss,
        rand_pub: ImageValue,
        rand_share: ImageValue,
        rand_elg_a: ImageValue,
        rand_id: ImageValue,
    ) -> ReencryptedChallenge:
        """
        """
        return cls(
            pvss,
            _asn1.ReencryptedChallenge(
                {
                    "parameters": pvss.params.asn1,
                    "public_keys": [share.pub.asn1 for share in pvss.shares.shares],
                    "shares": pvss.shares.asn1,
                    "receiver_public_key": pvss.receiver_public_key.asn1,
                    "rand_pub": rand_pub.asn1,
                    "rand_share": rand_share.asn1,
                    "rand_elg_a": rand_elg_a.asn1,
                    "rand_id": rand_id.asn1,
                }
            ),
        )


class Pvss:
    """
    Main class to work with Pvss. Stores all public messages and exposes the PVSS operations.
    """

    _params: SystemParameters
    _user_public_keys: Dict[str, PublicKey]
    _shares: SharedSecret
    _reencrypted_shares: List[ReencryptedShare]
    _receiver_public_key: PublicKey

    def __init__(self) -> None:
        """
        The constructor takes no parameters.
        """

        self._user_public_keys = {}
        self._reencrypted_shares = []

    @property
    def params(self) -> SystemParameters:
        """
        Retrieve system parameters.

        Returns:
            The system parameters.
        """

        return self._params

    def set_params(self, data: ByteString) -> SystemParameters:
        """
        Set system parameters.

        Args
            data: DER encoded system parameters.

        Returns:
            Decoded system parameters.

        Raises:
            Exception: If already set.
        """

        params = SystemParameters.from_der(self, data)
        if hasattr(self, "_params"):
            raise Exception("Parameters already set")
        self._params = params
        return params

    @property
    def user_public_keys(self) -> Dict[str, PublicKey]:
        """
        Retrieve all user public keys, as mapping from username to PublicKey.

        Returns:
            Mapping of username to PublicKey.
        """

        return dict(self._user_public_keys)

    def add_user_public_key(self, data: ByteString) -> PublicKey:
        """
        Add a user public key to the internal state.

        Args:
            data: DER encoded public key

        Returns:
            Decoded user public key.

        Raises:
            ValueError: On duplicate name or public key value
        """
        pub_key = PublicKey.from_der(self, data)

        for pub in self._user_public_keys.values():
            if pub_key.name == pub.name:
                raise ValueError(f"Duplicate name: {pub_key.name}")
            if pub_key.pub == pub.pub:
                raise ValueError(
                    f"Duplicate public key value in keys {pub_key.name} and {pub.name}"
                )
        self._user_public_keys[pub_key.name] = pub_key
        return pub_key

    @property
    def shares(self) -> SharedSecret:
        """
        Retrieve the shares of the secret.

        Returns:
            Shares of the secret.
        """

        return self._shares

    def set_shares(self, data: ByteString) -> SharedSecret:
        """
        Set the shares of the secret.

        Args:
            data: DER encoded secret shares.

        Returns:
            Decoded secret shares.

        Raises:
            Exception: If already set.
        """

        shares = SharedSecret.from_der(self, data)
        if hasattr(self, "_shares"):
            raise Exception("Shares already set")
        self._shares = shares
        return shares

    @property
    def reencrypted_shares(self) -> List[ReencryptedShare]:
        """
        Retrieve the list of reencrypted shares.

        Returns:
            List of reencrypted shares.
        """

        return self._reencrypted_shares

    def add_reencrypted_share(self, data: ByteString) -> ReencryptedShare:
        """
        Add a re-encrypted share to the internal state.

        Args:
            data: DER encoded re-encrypted share.

        Returns:
            Decoded reencrypted share.

        Raises:
            ValueError: On duplicate
        """
        reenc_share = ReencryptedShare.from_der(self, data)
        if reenc_share.idx in {s.idx for s in self._reencrypted_shares}:
            raise ValueError(f"Duplicate index: {reenc_share.idx}")
        self._reencrypted_shares.append(reenc_share)
        return reenc_share

    @property
    def receiver_public_key(self) -> PublicKey:
        """
        Retrieve receiver's public key.

        Returns:
            Receiver's public key.
        """

        return self._receiver_public_key

    def set_receiver_public_key(self, data: ByteString) -> PublicKey:
        """
        Add the receiver's public key to the internal state.

        Args:
            data: DER encoded receiver's public key.

        Returns:
            Decoded receiver's public key.

        Raises:
            Exception: On duplicate
        """

        pub = PublicKey.from_der(self, data)
        if hasattr(self, "_receiver_public_key"):
            raise Exception("Receiver key already set")
        self._receiver_public_key = pub
        return pub

    def create_user_keypair(self, name: str) -> Tuple[bytes, bytes]:
        """
        Create a random key pair for a user.

        Args:
            name: Name of key; will be included in the public key.

        Returns:
            DER encoded private key and public key
        """
        priv = PrivateKey.create_random(self)
        pub = priv.pub(name).der
        self.add_user_public_key(pub)
        return priv.der, pub

    def create_receiver_keypair(self, name: str) -> Tuple[bytes, bytes]:
        """
        Create a random key pair for the receiver.

        Args:
            name: Name of key; will be included in the public key.

        Returns:
            DER encoded private key and public key
        """
        priv = PrivateKey.create_random(self)
        pub = priv.pub(name).der
        self.set_receiver_public_key(pub)
        return priv.der, pub

    def share_secret(self, qualified_size: int) -> Tuple[bytes, bytes]:
        """
        Create a secret, split it and compute the encrypted shares.

        Args:
            qualified_size: Number of shares required to reconstruct the secret

        Returns:
            DER encoded shared secret and the DER encoded encrypted shares
        """

        secret, shares = SharedSecret.create_shared_secret(self, qualified_size)
        self.set_shares(shares.der)
        return secret.der, shares.der

    def reencrypt_share(self, der_private_key: ByteString) -> bytes:
        """
        Decrypt a share of the encrypted secret with the private_key and
        re-encrypt it with another public key

        Args:
            der_private_key: A user's DER encoded private key

        Returns:
            DER encoded re-encrypted share
        """

        private_key = PrivateKey.from_der(self, der_private_key)
        share = ReencryptedShare.reencrypt(self, private_key).der
        self.add_reencrypted_share(share)
        return share

    def reconstruct_secret(self, der_private_key: ByteString) -> bytes:
        """
        Decrypt the re-encrypted shares with the private key and reconstruct the secret

        Args:
            der_private_key: Receiver's DER encoded private key

        Returns:
            DER encoded secret
        """

        private_key = PrivateKey.from_der(self, der_private_key)
        return Secret.reconstruct(self, private_key).der


class Poly(Sequence[PreGroupValue]):
    """
    Polynomial
    """

    _coeffs: List[PreGroupValue]
    _zero: PreGroupValue

    def __init__(self, coeffs: Iterable[PreGroupValue], zero: PreGroupValue) -> None:
        self._coeffs = list(coeffs)
        self._zero = zero

    def __call__(self, i: int) -> PreGroupValue:
        return sum((coeff * (i ** j) for j, coeff in enumerate(self._coeffs)), self._zero)

    def __len__(self) -> int:
        return len(self._coeffs)

    @overload
    def __getitem__(self, index: int) -> PreGroupValue:
        """
        """

    @overload
    def __getitem__(self, index: slice) -> Sequence[PreGroupValue]:
        """
        """

    def __getitem__(
        self, idx: Union[int, slice]
    ) -> Union[PreGroupValue, Sequence[PreGroupValue]]:
        return self._coeffs[idx]

    def __repr__(self) -> str:
        return "Poly([" + ", ".join(str(coeff) for coeff in self._coeffs) + "])"



"""
Command line utility for PVSS
"""

from __future__ import annotations

import logging
from contextlib import suppress
from io import BufferedIOBase
from os import getcwd, umask
from pathlib import Path
from random import randrange

import click

from .pvss import Pvss, PublicKey


def pvss_from_datadir(datadir: Path) -> Pvss:
    """
    Read in the data directory; create it if it doesn't exist.
    """

    pvss = Pvss()

    # Read files in order. If any is missing (e.g. not generated yet), that's not a problem.
    with suppress(FileNotFoundError):
        pvss.set_params((datadir / "parameters").read_bytes())
        logging.info("Loaded parameters")

        for path in (datadir / "users").iterdir():
            pub = pvss.add_user_public_key(path.read_bytes())
            logging.info(f"Loaded user public key {path} ({pub.name})")

        pvss.set_shares((datadir / "shares").read_bytes())
        logging.info("Loaded shares")

        pvss.set_receiver_public_key((datadir / "receiver").read_bytes())
        logging.info("Loaded receiver public key")

        for path in (datadir / "reencrypted").iterdir():
            reenc_share = pvss.add_reencrypted_share(path.read_bytes())
            logging.info(f"Loaded reencrypted share {path} ({reenc_share.share.pub.name})")

    return pvss


def write_private(path: Path, contents: bytes) -> None:
    prev = umask(0o077)
    try:
        with path.open("xb") as fp:
            fp.write(contents)
    finally:
        umask(prev)


def write_public(path: Path, contents: bytes) -> None:
    with path.open("xb") as fp:
        fp.write(contents)


def write_public_random(path: Path, contents: bytes) -> None:
    path.mkdir(exist_ok=True)
    for __ in range(10):
        name = format(randrange(2 ** 32), "08x")
        with suppress(FileExistsError), (path / name).open("xb") as fp:
            fp.write(contents)
            return
    raise FileExistsError("Cannot find unused random name")  # pragma: no cover


@click.group()
@click.argument(
    "datadir",
    type=click.Path(file_okay=False, resolve_path=True),
    callback=lambda ctx, param, value: Path(value),
    required=True,
    metavar="DATADIR",
)
@click.pass_context
def cli(ctx: click.Context, datadir: Path) -> None:
    """
    Publicly Verifiable Secret Splitting

    PVSS is a cryptographic protocol to split a random secret amongst N
    users and T of those can later cooperate to reconstruct the secret.

    All messages can be exchanged over public channels and can be verified by
    everyone to be correct.

    https://en.wikipedia.org/wiki/Publicly_Verifiable_Secret_Sharing

    DATADIR needs to point at a directory which contains all the public
    messages that are being exchanged. If the directory does not exist, it is
    created.

    All sub commands read and verify all messages from that directory and add
    their own messages.
    """
    ctx.obj = datadir

    logging.basicConfig(level=logging.INFO)


@cli.group()
def genparams() -> None:
    """
    Generate system parameters.

    They are written to {datadir}/parameters.
    """


@genparams.command("qr")
@click.argument("dhparams", metavar="DHPARAMS", type=click.File(mode="rb"))
@click.pass_context
def genparams_qr(ctx: click.Context, dhparams: BufferedIOBase) -> None:
    """
    Generate QR system parameters.

    DHPARAMS is a file with diffie hellman parameters as created by
    `openssl dhparam 4096`, either DER or PEM format.

    PVSS works with almost any bit length, but the security level depends on it.
    4096 should be a sane choice.
    """
    from .qr import create_qr_params

    pvss = pvss_from_datadir(ctx.obj)
    params = create_qr_params(pvss, dhparams.read())

    ctx.obj.mkdir(exist_ok=True)
    write_public(ctx.obj / "parameters", params)


@genparams.command("rst255")
@click.pass_context
def genparams_rst255(ctx: click.Context) -> None:
    """
    Generate Ristretto255 system parameters.

    For details about Ristretto255, refer to https://ristretto.group/
    """
    from .ristretto_255 import create_ristretto_255_parameters

    pvss = pvss_from_datadir(ctx.obj)
    params = create_ristretto_255_parameters(pvss)

    ctx.obj.mkdir(exist_ok=True)
    write_public(ctx.obj / "parameters", params)


@cli.command()
@click.argument("name", metavar="NAME")
@click.argument(
    "keyfile",
    metavar="KEYFILE",
    type=click.Path(dir_okay=False, resolve_path=True),
    callback=lambda ctx, param, value: Path(value),
)
@click.pass_context
def genuser(ctx: click.Context, name: str, keyfile: Path) -> None:
    """
    Generate keypair for a new user.

    The NAME of the key owner is stored in the public key.
    The public key is written to {datadir}/users/ with random filename.

    The private key is written to KEYFILE.
    """
    pvss = pvss_from_datadir(ctx.obj)
    priv, pub = pvss.create_user_keypair(name)
    write_private(keyfile, priv)
    write_public_random(ctx.obj / "users", pub)


@cli.command()
@click.argument(
    "keyfile",
    metavar="KEYFILE",
    type=click.Path(dir_okay=False, resolve_path=True),
    callback=lambda ctx, param, value: Path(value),
)
@click.option("--name", help="Name of receiver", default="receiver")
@click.pass_context
def genreceiver(ctx: click.Context, keyfile: Path, name: str) -> None:
    """
    Generate keypair for the receiver.

    The private key is written to KEYFILE, the public key is written to {datadir}/receiver.
    """
    pvss = pvss_from_datadir(ctx.obj)
    priv, pub = pvss.create_receiver_keypair(name)
    write_private(keyfile, priv)
    write_public(ctx.obj / "receiver", pub)


@cli.command()
@click.argument("min_shares", metavar="MIN_SHARES", type=int)
@click.argument(
    "secretfile",
    metavar="SECRETFILE",
    type=click.Path(dir_okay=False, resolve_path=True),
    callback=lambda ctx, param, value: Path(value),
)
@click.pass_context
def splitsecret(ctx: click.Context, min_shares: int, secretfile: Path) -> None:
    """
    Generate and split random secret.

    This generates a new random secret and writes it into SECRETFILE.
    The secret is split it into multiple shares, each encrypted with a user's public key.
    Those shared are written to {datadir}/shares.

    At least MIN_SHARES users need to cooperate to reconstruct the secret.
    """
    pvss = pvss_from_datadir(ctx.obj)
    secret, shares = pvss.share_secret(min_shares)
    write_private(secretfile, secret)
    write_public(ctx.obj / "shares", shares)


@cli.command()
@click.argument("keyfile", metavar="KEYFILE", type=click.File(mode="rb"))
@click.pass_context
def reencrypt(ctx: click.Context, keyfile: BufferedIOBase) -> None:
    """
    Re-encrypt share with receiver key.

    Decrypt a user's share with their private key KEYFILE and re-encrypt it with the receiver's
    public key.  The re-encrypted share is written to {datadir}/reencrypted/{username}.
    """
    pvss = pvss_from_datadir(ctx.obj)
    reencrypted_share = pvss.reencrypt_share(keyfile.read())
    write_public_random(ctx.obj / "reencrypted", reencrypted_share)


@cli.command()
@click.argument("keyfile", metavar="KEYFILE", type=click.File(mode="rb"))
@click.argument(
    "secretfile",
    metavar="SECRETFILE",
    type=click.Path(dir_okay=False, resolve_path=True),
    callback=lambda ctx, param, value: Path(value),
)
@click.pass_context
def reconstruct(ctx: click.Context, keyfile: BufferedIOBase, secretfile: Path) -> None:
    """
    Reconstruct the secret.

    Decrypt re-encrypted shares with the receiver's private key KEYFILE and
    join the shares to reconstruct the secret.  It is written into SECRETFILE.
    """
    pvss = pvss_from_datadir(ctx.obj)
    secret = pvss.reconstruct_secret(keyfile.read())
    write_private(secretfile, secret)



from __future__ import annotations

from typing import Any as _Any
from typing import ByteString, Dict, Type, TypeVar, cast

from asn1crypto.core import (
    Any,
    Asn1Value,
    Choice,
    Integer,
    Null,
    ObjectIdentifier,
    OctetString,
    Sequence,
    SequenceOf,
    UTF8String,
)

_T = TypeVar("_T", bound="VerifiedLoader")


class VerifiedLoader(Asn1Value):
    @classmethod
    def load(
        cls: Type[_T], encoded_data: ByteString, strict: bool = False, **kwargs: Dict[str, _Any]
    ) -> _T:
        self = super().load(encoded_data, strict=True, **kwargs)

        if self.dump(True) != encoded_data:
            raise ValueError("Does not encode back to original")

        return self


# fmt: off
class PreGroupValue(Integer):
    """
    """


class ImgGroupValue(Choice):
    _alternatives = [
        ("QrValue", Integer),
        ("ECPoint", OctetString),
    ]


class PvssAlgorithmId(ObjectIdentifier):
    _base_oid = "1.3.6.1.4.1.55040.1.0.1"
    _map = {
        _base_oid + ".0": "qr_mod_p",
        _base_oid + ".1": "ristretto_255",
        # _base_oid + ".2": "ristretto_448", (or decaf?)
        # _base_oid + ".3": "nist_p_256", (or decaf?)
    }


class SystemParameters(VerifiedLoader, Sequence):
    _fields = [
        ("algorithm", PvssAlgorithmId),
        ("parameters", Any),
    ]

    _oid_pair = ("algorithm", "parameters")
    _oid_specs = {
        "qr_mod_p": Integer,
        "ristretto_255": Null,
        # "ristretto_448": Null,
        # "nist_p_256": Null,
    }


class PublicKey(VerifiedLoader, Sequence):
    _fields = [
        ("name", UTF8String),
        ("pub0", ImgGroupValue),
        ("pub1", ImgGroupValue),
    ]


class PrivateKey(VerifiedLoader, Sequence):
    _fields = [
        ("priv", PreGroupValue),
    ]


class Secret(VerifiedLoader, Sequence):
    _fields = [
        ("secret", ImgGroupValue),
    ]


class Share(Sequence):
    _fields = [
        ("pub", UTF8String),
        ("share", ImgGroupValue),
        ("response_f0", PreGroupValue),
        ("response_f1", PreGroupValue),
    ]


class Shares(SequenceOf):
    _child_spec = Share


class Coefficients(SequenceOf):
    _child_spec = ImgGroupValue


class SharedSecret(VerifiedLoader, Sequence):
    _fields = [
        ("shares", Shares),
        ("coefficients", Coefficients),
        ("challenge", OctetString),
    ]


class HashInputUser(Sequence):
    _fields = [
        ("pub", PublicKey),
        ("commitment", ImgGroupValue),
        ("random_commitment", ImgGroupValue),
        ("share", ImgGroupValue),
        ("random_share", ImgGroupValue),
    ]


class HashInputUsers(SequenceOf):
    _child_spec = HashInputUser


class SharesChallenge(Sequence):
    _fields = [
        ("parameters", SystemParameters),
        ("coefficients", Coefficients),
        ("users", HashInputUsers),
    ]


class PublicKeys(SequenceOf):
    _child_spec = PublicKey


class ReencryptedChallenge(Sequence):
    _fields = [
        ("parameters", SystemParameters),
        ("public_keys", PublicKeys),
        ("shares", SharedSecret),
        ("receiver_public_key", PublicKey),
        ("rand_pub", ImgGroupValue),
        ("rand_share", ImgGroupValue),
        ("rand_elg_a", ImgGroupValue),
        ("rand_id", ImgGroupValue),
    ]


class ReencryptedShare(VerifiedLoader, Sequence):
    _fields = [
        ("idx", Integer),
        ("elg_a", ImgGroupValue),
        ("elg_b", ImgGroupValue),
        ("response_priv", PreGroupValue),
        ("response_v0", PreGroupValue),
        ("response_v1", PreGroupValue),
        ("response_w0", PreGroupValue),
        ("response_w1", PreGroupValue),
        ("challenge", OctetString),
    ]


class PvssContainer(Choice):
    _alternatives = [
        ("parameters", SystemParameters, {"explicit": 0}),
        ("priv_key", PrivateKey, {"explicit": 1}),
        ("user_pub", PublicKey, {"explicit": 2}),
        ("recv_pub", PublicKey, {"explicit": 3}),
        ("shared_secret", SharedSecret, {"explicit": 4}),
        ("reencrypted_share", ReencryptedShare, {"explicit": 5}),
    ]
# fmt: on



"""
Quadratic residues modulo safe prime p
"""

from __future__ import annotations

import hmac
from dataclasses import dataclass
from fractions import Fraction
from os import environ
from typing import TYPE_CHECKING, ByteString, Optional, Union, cast

from asn1crypto.algos import DHParameters
from asn1crypto.core import Asn1Value, Integer
from asn1crypto.pem import unarmor

try:
    from gmpy2 import invert, is_prime, legendre, mpz, powmod
except ImportError:
    # Work around the fact that gmpy2 is not installed in the readthedocs build image
    if "READTHEDOCS" not in environ:
        raise
else:
    # Will be fixed by gmpy2 2.1
    mpz_type = mpz if isinstance(mpz, type) else type(mpz(0))

from .asn1 import ImgGroupValue
from .groups import ImageGroup, ImageValue, PgvOrInt
from .pvss import Pvss, SystemParameters
from .zq import ZqGroup, ZqValue

if TYPE_CHECKING:  # pragma: no cover
    lazy = property
else:
    from lazy import lazy


def create_qr_params(pvss: Pvss, params: Union[int, str, ByteString]) -> bytes:
    """
    Create and set QR parameters.

    If params is str or a ByteString, assume it's a diffie-hellman parameter file
    such as created by "openssl dhparam 4096", either DER or PEM encoded.

    Args:
        pvss: Pvss object with public values
        params: if int, must be a safe prime, otherwise must be a DH params file with a safe prime.

    Returns:
        DER encoded QR system parameters.
    """

    if not isinstance(params, int):
        if isinstance(params, str):
            params = params.encode()

        if params[0] != 0x30:
            # DER encoding of "Sequence" starts with 0x30. So it's probably PEM.
            params = unarmor(bytes(params))[2]

        params = int(cast(Integer, DHParameters.load(bytes(params))["p"]))

    result = QrParameters.create(pvss, params).der
    pvss.set_params(result)
    return result


class QrParameters(SystemParameters):
    """
    Quadratic residues modulo p.
    """

    ALGO = "qr_mod_p"

    @lazy
    def pre_group(self) -> ZqGroup:
        return ZqGroup(mpz(self.img_group.len))

    @lazy
    def img_group(self) -> QrGroup:
        """
        Create image group singleton
        """

        return QrGroup(mpz(int(cast(Integer, self.asn1["parameters"]))))

    def _make_gen(self, seed: str) -> QrValue:
        """
        Deterministic pseudo-random function to generate a group element.
        """
        key = seed.encode()
        p = self.img_group.p
        bits = p.bit_length() * 2

        dig = self.der
        tmp = bytearray()

        while len(tmp) * 8 < bits:
            dig = hmac.digest(key, dig, "sha256")
            tmp.extend(dig)

        # Square to get a quadratic residue
        value = powmod(int.from_bytes(tmp, "big"), 2, p)

        # Check for bad value. 0 is no group element at all. 1 is the neutral element.
        if value in {0, 1}:
            # Try again with new seed. This happening is extremely unlikely.
            return self._make_gen(seed + "_")

        return self.img_group(value)


@dataclass(frozen=True)
class QrGroup(ImageGroup):
    """
    Multiplicative group of Quadratic Residues modulo safe prime p
    """

    p: mpz

    def __post_init__(self) -> None:
        if not is_prime(self.p):
            raise ValueError("p not prime")

        if not is_prime(self.len):
            raise ValueError("(p - 1) / 2 not prime")

    def __call__(self, value: Union[int, mpz, Integer, Asn1Value]) -> QrValue:
        if isinstance(value, (int, mpz_type)):
            value %= self.p
            if value == 0:
                raise ValueError("0 not in group")
            return QrValue(self, value)

        if isinstance(value, ImgGroupValue):
            value = value.chosen

        if not isinstance(value, Integer):
            raise TypeError(type(value))

        value = mpz(int(value))

        if not 1 <= value < self.p:
            raise ValueError("Not a valid group element")

        if legendre(value, self.p) != 1:
            raise ValueError("Not a valid group element")

        return QrValue(self, value)

    @lazy
    def len(self) -> int:
        return int(self.p // 2)

    def __repr__(self) -> str:
        """
        Outputs a representation of this group.

        Returns:
            Representation of this group
        """

        return f"QrGroup({self.p})"


@dataclass(frozen=True)
class QrValue(ImageValue):
    """
    Elements of QrGroup. The value is always in [1, p).
    """

    group: QrGroup
    value: mpz

    @property
    def asn1(self) -> ImgGroupValue:
        """
        Convert to ASN.1 Integer type
        """

        return ImgGroupValue({"QrValue": Integer(int(self))})

    def __int__(self) -> int:
        """
        Returns:
            Value of self as a native python integer
        """
        return int(self.value)

    def __mul__(self, other: ImageValue) -> QrValue:
        """
        Implement the group operation a * b.

        Args:
            other: Second operand

        Returns:
            Product of `self` and `other`
        """

        if not isinstance(other, QrValue):
            return NotImplemented

        return self.group(self.value * other.value)

    def __pow__(
        self, other: Union[PgvOrInt, Fraction], modulo: Optional[int] = None
    ) -> QrValue:
        """
        Implement a ** b and pow(a, b).
        If b is a Fraction c/d, compute a ** (c * (d^-1))

        Args:
            other: exponent
            modulo: Not supported, must be None

        Returns:
            `self` raised to the power of `other`
        """

        if modulo is not None:
            raise Exception("Modulo not supported")

        if isinstance(other, int):
            return self.group(powmod(self.value, other, self.group.p))

        if isinstance(other, ZqValue):
            if other.group.len != self.group.len:
                raise TypeError("incompatible groups")
            return self.group(powmod(self.value, int(other), self.group.p))

        if isinstance(other, Fraction):
            tmp = powmod(self.value, other.numerator, self.group.p)

            if other.denominator != 1:
                tmp = powmod(tmp, invert(other.denominator, self.group.len), self.group.p)
            return self.group(tmp)

        return NotImplemented

    def __repr__(self) -> str:
        """
        Outputs a representation of this value.

        Returns:
            Representation of this value
        """

        return f"{self.group}({int(self)})"



"""
Binding to `libsodium <https://libsodium.org/>`_ to use `Ristretto255 <https://ristretto.group/>`_ for group operations.
"""

from __future__ import annotations

import ctypes
import ctypes.util
import hmac
from dataclasses import dataclass
from fractions import Fraction
from os import environ
from secrets import randbelow
from typing import TYPE_CHECKING, ByteString, Optional, Union

from asn1crypto.core import Asn1Value, Integer, OctetString

from . import asn1 as _asn1
from .groups import ImageGroup, ImageValue, PgvOrInt, PreGroup, PreGroupValue
from .pvss import Pvss, SystemParameters

if TYPE_CHECKING:  # pragma: no cover
    lazy = property
else:
    from lazy import lazy


# Order of the Ristretto255 group.
group_order = 2 ** 252 + 27742317777372353535851937790883648493


def create_ristretto_255_parameters(pvss: Pvss) -> bytes:
    """
    Create and set Ristretto255 parameters.

    Args:
        pvss: Pvss object with public values

    Returns:
        DER encoded Ristretto255 system parameters.
    """

    result = Ristretto255Parameters.create(pvss, None).der
    pvss.set_params(result)
    return result


class Ristretto255Parameters(SystemParameters):
    ALGO = "ristretto_255"

    @lazy
    def pre_group(self) -> Ristretto255ScalarGroup:
        return Ristretto255ScalarGroup()

    @lazy
    def img_group(self) -> Ristretto255Group:
        return Ristretto255Group(self.pre_group)

    def _make_gen(self, seed: str) -> Ristretto255Point:
        while True:
            dig = hmac.digest(seed.encode(), self.der, "sha512")
            gen = self.img_group.from_hash(dig)

            # Check for neutral elements; we don't want those.
            if gen:
                return gen
            else:  # pragma: no cover
                # Try again with other seed.
                seed += "_"


class _Lib:
    try:
        lib_name = ctypes.util.find_library("sodium")
        if not lib_name:  # pragma: no cover
            raise Exception("libsodium not found")

        lib = ctypes.cdll.LoadLibrary(lib_name)
        if lib.sodium_init() < 0:  # pragma: no cover
            raise Exception("Cannot initialize libsodium")

        # int sodium_memcmp(const void * const b1_, const void * const b2_, size_t len);
        memcmp = lib.sodium_memcmp
        memcmp.restype = ctypes.c_int
        memcmp.argtypes = ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t

        # int sodium_is_zero(const unsigned char *n, const size_t nlen);
        is_zero = lib.sodium_is_zero
        is_zero.restype = ctypes.c_int
        is_zero.argtypes = ctypes.c_char_p, ctypes.c_size_t

        # int crypto_core_ristretto255_is_valid_point(const unsigned char *p);
        point_is_valid = lib.crypto_core_ristretto255_is_valid_point
        point_is_valid.restype = ctypes.c_int
        point_is_valid.argtypes = (ctypes.c_char_p,)

        # void crypto_core_ristretto255_random(unsigned char *p);
        point_random = lib.crypto_core_ristretto255_random
        point_random.restype = None
        point_random.argtypes = (ctypes.c_char_p,)

        # int crypto_core_ristretto255_from_hash(unsigned char *p, const unsigned char *r);
        point_from_hash = lib.crypto_core_ristretto255_from_hash
        point_from_hash.restype = ctypes.c_int
        point_from_hash.argtypes = ctypes.c_char_p, ctypes.c_char_p

        # int crypto_scalarmult_ristretto255(unsigned char *q, const unsigned char *n, const unsigned char *p);
        point_mul = lib.crypto_scalarmult_ristretto255
        point_mul.restype = ctypes.c_int
        point_mul.argtypes = ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p

        # int crypto_scalarmult_ristretto255_base(unsigned char *q, const unsigned char *n);
        point_base_mul = lib.crypto_scalarmult_ristretto255_base
        point_base_mul.restype = ctypes.c_int
        point_base_mul.argtypes = ctypes.c_char_p, ctypes.c_char_p

        # int crypto_core_ristretto255_add(unsigned char *r, const unsigned char *p, const unsigned char *q);
        point_add = lib.crypto_core_ristretto255_add
        point_add.restype = ctypes.c_int
        point_add.argtypes = ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p

        # int crypto_core_ristretto255_sub(unsigned char *r, const unsigned char *p, const unsigned char *q);
        point_sub = lib.crypto_core_ristretto255_sub
        point_sub.restype = ctypes.c_int
        point_sub.argtypes = ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p

        # void crypto_core_ristretto255_scalar_random(unsigned char *r);
        scalar_random = lib.crypto_core_ristretto255_scalar_random
        scalar_random.restype = None
        scalar_random.argtypes = (ctypes.c_char_p,)

        # void crypto_core_ristretto255_scalar_reduce(unsigned char *r, const unsigned char *s);
        scalar_reduce = lib.crypto_core_ristretto255_scalar_reduce
        scalar_reduce.restype = None
        scalar_reduce.argtypes = ctypes.c_char_p, ctypes.c_char_p

        # int crypto_core_ristretto255_scalar_invert(unsigned char *recip, const unsigned char *s);
        scalar_invert = lib.crypto_core_ristretto255_scalar_invert
        scalar_invert.restype = ctypes.c_int
        scalar_invert.argtypes = ctypes.c_char_p, ctypes.c_char_p

        # void crypto_core_ristretto255_scalar_negate(unsigned char *neg, const unsigned char *s);
        scalar_negate = lib.crypto_core_ristretto255_scalar_negate
        scalar_negate.restype = None
        scalar_negate.argtypes = ctypes.c_char_p, ctypes.c_char_p

        # void crypto_core_ristretto255_scalar_complement(unsigned char *comp, const unsigned char *s);
        scalar_complement = lib.crypto_core_ristretto255_scalar_complement
        scalar_complement.restype = None
        scalar_complement.argtypes = ctypes.c_char_p, ctypes.c_char_p

        # void crypto_core_ristretto255_scalar_add(unsigned char *z, const unsigned char *x, const unsigned char *y);
        scalar_add = lib.crypto_core_ristretto255_scalar_add
        scalar_add.restype = None
        scalar_add.argtypes = ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p

        # void crypto_core_ristretto255_scalar_sub(unsigned char *z, const unsigned char *x, const unsigned char *y);
        scalar_sub = lib.crypto_core_ristretto255_scalar_sub
        scalar_sub.restype = None
        scalar_sub.argtypes = ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p

        # void crypto_core_ristretto255_scalar_mul(unsigned char *z, const unsigned char *x, const unsigned char *y);
        scalar_mul = lib.crypto_core_ristretto255_scalar_mul
        scalar_mul.restype = None
        scalar_mul.argtypes = ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p

    except Exception:
        # Work around the fact that libsodium is not installed in the readthedocs build image
        if "READTHEDOCS" not in environ:
            raise


@dataclass(frozen=True, eq=False, repr=False)
class Ristretto255Group(ImageGroup):
    pre_group: Ristretto255ScalarGroup

    def __call__(self, value: Union[Asn1Value]) -> Ristretto255Point:
        """
        Create from serialized buffer, inverse of bytes()
        """
        if not isinstance(value, _asn1.ImgGroupValue):
            raise TypeError(type(value))
        buf = value.chosen
        if not isinstance(buf, OctetString):
            raise TypeError(type(buf))

        res = ctypes.create_string_buffer(bytes(buf), 32)
        if not _Lib.point_is_valid(res):
            raise ValueError("Not a valid point")
        return Ristretto255Point(self, res)

    def random(self) -> Ristretto255Point:
        """
        Generate random element.
        """

        res = ctypes.create_string_buffer(32)
        _Lib.point_random(res)
        return Ristretto255Point(self, res)

    def from_hash(self, value: ByteString) -> Ristretto255Point:
        """
        Generate a point from from up to 64 bytes. Those would usually come out of a hash function.
        """

        buf = ctypes.create_string_buffer(bytes(value), 64)
        res = ctypes.create_string_buffer(32)
        if _Lib.point_from_hash(res, buf) < 0:
            raise Exception("Unknown error")  # pragma: no cover
        return Ristretto255Point(self, res)

    @property
    def len(self) -> int:
        return group_order

    def __repr__(self) -> str:
        return "Ristretto255Group()"


@dataclass(frozen=True, eq=False, repr=False)
class Ristretto255Point(ImageValue):
    group: Ristretto255Group
    _buf: ctypes.Array[ctypes.c_char]

    @lazy
    def asn1(self) -> _asn1.ImgGroupValue:
        return _asn1.ImgGroupValue({"ECPoint": OctetString(bytes(self))})

    def __pow__(
        self, other: Union[PgvOrInt, Fraction], modulo: Optional[int] = None
    ) -> Ristretto255Point:
        """
        Compute self ** other
        """

        if modulo is not None:
            raise TypeError("modulo must be None")

        if isinstance(other, (int, Fraction)):
            buf = self.group.pre_group(other)._buf
        elif isinstance(other, Ristretto255Scalar):
            buf = other._buf
        else:
            return NotImplemented

        res = ctypes.create_string_buffer(32)
        if _Lib.point_mul(res, buf, self._buf) < 0:
            raise ValueError("Zero")
        return Ristretto255Point(self.group, res)

    def __mul__(self, other: ImageValue) -> Ristretto255Point:
        """
        Compute self * other
        """

        if not isinstance(other, Ristretto255Point):
            return NotImplemented

        res = ctypes.create_string_buffer(32)
        if _Lib.point_add(res, self._buf, other._buf) < 0:
            raise ValueError("Encoding error")
        return Ristretto255Point(self.group, res)

    def __floordiv__(self, other: ImageValue) -> Ristretto255Point:
        """
        Compute self / other
        """

        if not isinstance(other, Ristretto255Point):
            return NotImplemented

        res = ctypes.create_string_buffer(32)
        if _Lib.point_sub(res, self._buf, other._buf) < 0:
            raise ValueError("Encoding error")
        return Ristretto255Point(self.group, res)

    def __bytes__(self) -> bytes:
        """
        """

        return bytes(self._buf)

    def __str__(self) -> str:
        """
        """

        return f"Ristretto255Point(0x{bytes(self).hex()})"

    def __repr__(self) -> str:
        """
        """

        return f"Ristretto255Point.from_bytes({bytes(self)!r})"

    def __eq__(self, other: object) -> bool:
        """
        """

        if not isinstance(other, Ristretto255Point):
            return False

        return not _Lib.memcmp(self._buf, other._buf, 32)

    def __bool__(self) -> bool:
        """
        False if this is the neutral element
        """

        return not _Lib.is_zero(self._buf, 32)

    def __hash__(self) -> int:
        return hash(bytes(self._buf))


class Ristretto255ScalarGroup(PreGroup):
    def __call__(self, value: Union[int, Asn1Value, Fraction]) -> Ristretto255Scalar:
        """
        Convert an integer into a group element

        Returns:
            Group element
        """

        if isinstance(value, _asn1.PreGroupValue):
            value = int(value)
            if not 0 <= value < group_order:
                raise ValueError("Not a valid group element")
            res = ctypes.create_string_buffer(value.to_bytes(32, "little"), 32)
            return Ristretto255Scalar(self, res)

        if isinstance(value, int):
            if value < 0:
                neg = True
                value = -value
            else:
                neg = False

            val = ctypes.create_string_buffer(value.to_bytes(64, "little"), 64)
            res = ctypes.create_string_buffer(32)
            _Lib.scalar_reduce(res, val)

            if neg:
                return -Ristretto255Scalar(self, res)
            else:
                return Ristretto255Scalar(self, res)

        if isinstance(value, Fraction):
            return self(value.numerator) * self(value.denominator).inv

        raise TypeError(type(value))

    @property
    def len(self) -> int:
        """
        Get number of elements in this group

        Returns:
            group size
        """

        return group_order

    @property
    def rand(self) -> Ristretto255Scalar:
        """
        Create random element of this group

        Returns:
            Random group element
        """

        value = randbelow(group_order)
        res = ctypes.create_string_buffer(value.to_bytes(32, "little"), 32)
        return Ristretto255Scalar(self, res)

    @property
    def rand_nonzero(self) -> Ristretto255Scalar:
        """
        Create random element of this group, but never the neutral element.

        Returns:
            Random group element
        """

        res = ctypes.create_string_buffer(32)
        _Lib.scalar_random(res)
        return Ristretto255Scalar(self, res)

    def __repr__(self) -> str:
        """
        Outputs a representation of this group.

        Returns:
            Representation of this group
        """

        return "Ristretto255ScalarGroup()"


@dataclass(frozen=True, eq=False, repr=False)
class Ristretto255Scalar(PreGroupValue):
    group: Ristretto255ScalarGroup
    _buf: ctypes.Array[ctypes.c_char]

    def __neg__(self) -> Ristretto255Scalar:
        res = ctypes.create_string_buffer(32)
        _Lib.scalar_negate(res, self._buf)
        return Ristretto255Scalar(self.group, res)

    def __add__(self, other: Union[int, PreGroupValue]) -> Ristretto255Scalar:
        if isinstance(other, int):
            buf = self.group(other)._buf
        elif isinstance(other, Ristretto255Scalar):
            buf = other._buf
        else:
            return NotImplemented

        res = ctypes.create_string_buffer(32)
        _Lib.scalar_add(res, self._buf, buf)
        return Ristretto255Scalar(self.group, res)

    def __sub__(self, other: Union[int, PreGroupValue]) -> Ristretto255Scalar:
        if isinstance(other, int):
            buf = self.group(other)._buf
        elif isinstance(other, Ristretto255Scalar):
            buf = other._buf
        else:
            return NotImplemented

        res = ctypes.create_string_buffer(32)
        _Lib.scalar_sub(res, self._buf, buf)
        return Ristretto255Scalar(self.group, res)

    def __mul__(self, other: Union[int, PreGroupValue]) -> Ristretto255Scalar:
        if isinstance(other, int):
            buf = self.group(other)._buf
        elif isinstance(other, Ristretto255Scalar):
            buf = other._buf
        else:
            return NotImplemented

        res = ctypes.create_string_buffer(32)
        _Lib.scalar_mul(res, self._buf, buf)
        return Ristretto255Scalar(self.group, res)

    @lazy
    def inv(self) -> Ristretto255Scalar:
        res = ctypes.create_string_buffer(32)
        if _Lib.scalar_invert(res, self._buf) < 0:
            raise ValueError("Cannot invert value")

        return Ristretto255Scalar(self.group, res)

    def __eq__(self, other: object) -> bool:
        """
        """

        if not isinstance(other, Ristretto255Scalar):
            return False

        return not _Lib.memcmp(self._buf, other._buf, 32)

    def __bool__(self) -> bool:
        return not _Lib.is_zero(self._buf, 32)

    def __repr__(self) -> str:
        return f"Ristretto255ScalarGroup()({int(self)})"

    def __int__(self) -> int:
        return int.from_bytes(bytes(self._buf), "little")

    def __bytes__(self) -> bytes:
        return bytes(self._buf)

    @lazy
    def asn1(self) -> _asn1.PreGroupValue:
        return _asn1.PreGroupValue(int(self))

    def __hash__(self) -> int:
        return hash(bytes(self))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pvss']

package_data = \
{'': ['*']}

install_requires = \
['asn1crypto>=1.3.0,<2.0.0', 'click>=7.0,<8.0', 'lazy>=1.4,<2.0']

extras_require = \
{'docs': ['sphinx>=2.4,<3.0', 'sphinx-rtd-theme>=0.4,<0.5']}

entry_points = \
{'console_scripts': ['pvss = pvss.cli:cli']}

setup_kwargs = {
    'name': 'pvss',
    'version': '0.2.0',
    'description': 'Public Verifiable Secret Splitting in Python',
    'long_description': "Publicly Verifiable Secret Splitting in python\n==============================================\nThis project is a python (>= 3.7) implementation (library and CLI) of\n`Publicly Verifiable Secret Splitting (PVSS)\n<https://en.wikipedia.org/wiki/Publicly_Verifiable_Secret_Sharing>`_.\n\nPVSS is a non-interactive cryptographic protocol between multiple participants\nfor splitting a random secret into multiple shares and distributing them amongst a\ngroup of users.  An arbitrary subset of those users (e.g. any 3 out of 5) can\nlater cooperate to reassemble the secret.\n\nThe common use case for secret splitting is to create a highly durable backup of\nhighly sensitive data such as cryptographic keys.\n\nAll communication between the participants is public and everyone can verify\nthat all messages have been correctly created according to the protocol. This\nverification is done through `non-interactive zero-knowledge proofs\n<https://en.wikipedia.org/wiki/Non-interactive_zero-knowledge_proof>`_.\n\nThe math is based upon the paper `Non-Interactive and Information-Theoretic\nSecure Publicly Verifiable Secret Sharing <https://eprint.iacr.org/2004/201.ps>`_\nby *Chunming Tang* et al. who extended *Berry Schoenmaker*'s paper\n`A Simple Publicly Verifiable Secret Sharing Scheme and its Application to Electronic Voting\n<https://www.win.tue.nl/~berry/papers/crypto99.pdf>`_ which in turn is based on\n`Shamir's Secret Sharing <https://en.wikipedia.org/wiki/Shamir%27s_Secret_Sharing>`_.\n\nOne notable difference to prior work is the addition of a receiver user:\nIn their scheme the secret is made public while it is being reassembled, which\nviolates the goal to keep the secret secret. To address this issue, the users no longer\ndisclose their share of the secret but use `ElGamal encryption\n<https://en.wikipedia.org/wiki/ElGamal_encryption>`_ to securely convey the share to a\nseparate receiver user who will then reassemble the secret. Like all other communication,\nthe encrypted share is public and it can be verified that the users followed the protocol.\n\nDocumentation\n-------------\nFull documentation can be found at https://pvss.1e8.de/.\n\nInstallation\n------------\n``PVSS``'s dependencies are:\n\n* python (>= 3.7)\n* At least one of:\n    + `libsodium <https://libsodium.org/>`_ (>= 1.0.18, recommended, for `Ristretto255 <https://ristretto.group/>`_ group)\n\n      On Debian (Bullseye / 11 and later) or Ubuntu (Eoan / 19.10 and later):\n\n      .. code-block:: console\n\n          # apt install libsodium23\n\n    + `gmpy2 <https://pypi.org/project/gmpy2/>`_ (Group of quadratic residues modulo a large safe prime)\n\nYou can install ``PVSS`` with ``pip``:\n\n.. code-block:: console\n\n    $ pip install pvss\n\nAnd optionally:\n\n.. code-block:: console\n\n    $ pip install gmpy2\n\n\nExample\n-------\nThe following sequence of shell commands is executed by six different users who\nshare a data directory. E.g. use git to synchronize it between the users. All\nfiles inside ``datadir`` are public. All files outside of it are private.\n\n.. code-block:: console\n\n    (init)     $ pvss datadir genparams rst255 \n    (alice)    $ pvss datadir genuser Alice alice.key \n    (boris)    $ pvss datadir genuser Boris boris.key \n    (chris)    $ pvss datadir genuser Chris chris.key \n    (dealer)   $ pvss datadir splitsecret 2 secret0.der \n    (receiver) $ pvss datadir genreceiver recv.key \n    (boris)    $ pvss datadir reencrypt boris.key \n    (alice)    $ pvss datadir reencrypt alice.key \n    (receiver) $ pvss datadir reconstruct recv.key secret1.der \n\n``secret0.der`` and ``secret1.der`` should compare equal.\nThe *dealer* and *receiver* can encrypt an actual payload by using that file as a shared key.\n",
    'author': 'Jörn Heissler',
    'author_email': 'nosuchaddress@joern.heissler.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/joernheissler/pvss',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Implementation of pre-image and image groups for the group homomorphisms
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from fractions import Fraction
from typing import Optional, Union

from asn1crypto.core import Asn1Value

from . import asn1 as _asn1


class PreGroup(ABC):
    """
    Additive group of integers modulo a prime.
    """

    @abstractmethod
    def __call__(self, value: Union[int, Asn1Value]) -> PreGroupValue:
        """
        Convert an integer into a group element

        Returns:
            Group element
        """

    @property
    @abstractmethod
    def len(self) -> int:
        """
        Get number of elements in this group

        Returns:
            group size
        """

    @property
    @abstractmethod
    def rand(self) -> PreGroupValue:
        """
        Create random element of this group

        Returns:
            Random group element
        """

    @property
    @abstractmethod
    def rand_nonzero(self) -> PreGroupValue:
        """
        Create random element of this group, but never the neutral element.

        Returns:
            Random group element
        """

    @abstractmethod
    def __repr__(self) -> str:
        """
        Outputs a representation of this group.

        Returns:
            Representation of this group
        """


class PreGroupValue(ABC):
    group: PreGroup

    @abstractmethod
    def __int__(self) -> int:
        """
        Implement int(a)

        Returns:
            value
        """

    @abstractmethod
    def __neg__(self) -> PreGroupValue:
        """
        Implement -a

        Returns:
            inverse value
        """

    @abstractmethod
    def __add__(self, other: PgvOrInt) -> PreGroupValue:
        """
        Implement a + b

        Args:
            other: Second operand

        Returns:
            Sum of `self` and `other`
        """

    def __radd__(self, other: PgvOrInt) -> PreGroupValue:
        """
        Implement b + a

        Args:
            other: First operand

        Returns:
            Sum of `other` and `self`
        """

        return self.__add__(other)

    @abstractmethod
    def __mul__(self, other: PgvOrInt) -> PreGroupValue:
        """
        Implement a * b

        Args:
            other: Second operand

        Returns:
            Product of `self` and `other`
        """

    def __rmul__(self, other: PgvOrInt) -> PreGroupValue:
        """
        Implement b * a

        Args:
            other: First operand

        Returns:
            Product of `other` and `self`
        """

        return self.__mul__(other)

    @property
    @abstractmethod
    def inv(self) -> PreGroupValue:
        """
        Implement multiplicative inverse such that inv(x) * x == 1

        Returns:
            Multiplicative inverse of `self`
        """

    @abstractmethod
    def __repr__(self) -> str:
        """
        Outputs a representation of this value.

        Returns:
            Representation of this value
        """

    @property
    @abstractmethod
    def asn1(self) -> _asn1.PreGroupValue:
        """
        Convert value to an ASN.1 type so it can be serialized to DER.

        Returns:
            Value converted to an ASN.1 value
        """


PgvOrInt = Union[PreGroupValue, int]


class ImageGroup(ABC):
    """
    Abstract image group, can e.g. be some elliptic curve or quadratic residues modulo p
    """

    @abstractmethod
    def __call__(self, value: Union[Asn1Value]) -> ImageValue:
        """
        Create element of this group
        """

    @property
    @abstractmethod
    def len(self) -> int:
        """
        Get number of elements in this group

        Returns:
            group size
        """

    @abstractmethod
    def __repr__(self) -> str:
        """
        Outputs a representation of this group.

        Returns:
            Representation of this group
        """


class ImageValue(ABC):
    """
    Abstract image value, e.g. a curve point or a quadratic residue modulo p
    """

    group: ImageGroup

    @property
    @abstractmethod
    def asn1(self) -> _asn1.ImgGroupValue:
        """
        Convert value to an ASN.1 type so it can be serialized to DER.

        Returns:
            Value converted to an ASN.1 value
        """

    @abstractmethod
    def __mul__(self, other: ImageValue) -> ImageValue:
        """
        Implement the group operation a * b.

        Args:
            other: Second operand

        Returns:
            Product of `self` and `other`
        """

    @abstractmethod
    def __pow__(
        self, other: Union[PgvOrInt, Fraction], modulo: Optional[int] = None
    ) -> ImageValue:
        """
        Implement a ** b and pow(a, b), i.e. the repeated application of the group operation to `a`.

        Args:
            other: exponent
            modulo: Not supported, must be None

        Returns:
            `self` raised to the power of `other`
        """



"""
"""

from __future__ import annotations

from dataclasses import dataclass
from os import environ
from secrets import randbelow
from typing import TYPE_CHECKING, Any, Union, cast

from asn1crypto.core import Asn1Value

try:
    from gmpy2 import invert, is_prime, mpz
except ImportError:
    # Work around the fact that gmpy2 is not installed in the readthedocs build image
    if "READTHEDOCS" not in environ:
        raise

from . import asn1
from .groups import PgvOrInt, PreGroup, PreGroupValue

if TYPE_CHECKING:  # pragma: no cover
    lazy = property
else:
    from lazy import lazy


@dataclass(frozen=True)
class ZqGroup(PreGroup):
    """
    Additive group of integers modulo Sophie Germain prime q
    """

    q: mpz

    def __post_init__(self) -> None:
        """
        Ensure that q and 2q+1 are prime.
        """

        if self.q < 0:
            raise ValueError("q is negative")

        if not is_prime(self.q):
            raise ValueError("q not prime")

        if not is_prime(self.q * 2 + 1):
            raise ValueError("2q + 1 not prime")

    def __call__(self, value: Union[int, Asn1Value]) -> ZqValue:
        """
        Convert an integer into a group element

        Returns:
            Group element
        """

        if isinstance(value, int):
            return ZqValue(self, value % self.q)

        if isinstance(value, asn1.PreGroupValue):
            mpz_value = mpz(int(value))
            if not 0 <= mpz_value < self.q:
                raise ValueError("Not a valid group element")
            return ZqValue(self, mpz_value)

        raise TypeError(f"Type not supported: {type(value)}")

    @property
    def len(self) -> int:
        """
        Get number of elements in this group

        Returns:
            group size
        """

        return int(self.q)

    @property
    def rand(self) -> ZqValue:
        """
        Create random element of this group

        Returns:
            Random group element
        """

        return self(randbelow(int(self.q)))

    @property
    def rand_nonzero(self) -> ZqValue:
        """
        Create random element of this group, but never the neutral element.

        Returns:
            Random group element
        """

        return self(1 + randbelow(int(self.q - 1)))

    def __repr__(self) -> str:
        """
        Outputs a representation of this group.

        Returns:
            Representation of this group
        """

        return f"ZqGroup({self.q})"


@dataclass(frozen=True, eq=False)
class ZqValue(PreGroupValue):
    """
    Elements of ZqGroup
    """

    group: ZqGroup
    _value: mpz

    def __int__(self) -> int:
        """
        Implement int(a)

        Returns:
            value
        """

        return int(self._value)

    def __neg__(self) -> ZqValue:
        """
        Implement -a

        Returns:
            inverse value
        """

        return ZqValue(self.group, -self._value % self.group.q)

    def __add__(self, other: PgvOrInt) -> ZqValue:
        """
        Implement a + b

        Args:
            other: Second operand

        Returns:
            Sum of `self` and `other`
        """

        if isinstance(other, int):
            return ZqValue(self.group, (self._value + other) % self.group.q)

        if isinstance(other, ZqValue):
            if self.group is not other.group:
                raise TypeError("Group mismatch")

            return ZqValue(self.group, (self._value + other._value) % self.group.q)

        return NotImplemented

    def __mul__(self, other: PgvOrInt) -> ZqValue:
        """
        Implement a * b

        Args:
            other: Second operand

        Returns:
            Product of `self` and `other`
        """

        if isinstance(other, int):
            return ZqValue(self.group, self._value * other % self.group.q)

        if isinstance(other, ZqValue):
            if self.group is not other.group:
                raise TypeError("Group mismatch")

            return ZqValue(self.group, self._value * other._value % self.group.q)

        return NotImplemented

    @property
    def inv(self) -> ZqValue:
        """
        Implement multiplicative inverse such that inv(x) * x == 1

        Returns:
            Multiplicative inverse of `self`
        """
        return ZqValue(self.group, invert(self._value, self.group.q))

    def __repr__(self) -> str:
        """
        Outputs a representation of this value.

        Returns:
            Representation of this value
        """

        return f"{self.group}({self._value})"

    @lazy
    def asn1(self) -> asn1.PreGroupValue:
        """
        Convert value to an ASN.1 type so it can be serialized to DER.

        Returns:
            Value converted to an ASN.1 value
        """
        return asn1.PreGroupValue(int(self._value))

    def __eq__(self, other: Any) -> bool:
        """
        """

        if isinstance(other, ZqValue):
            # "is not" is by intention to force usage of the identical group.
            if self.group is not other.group:
                return False

            return self._value == other._value

        return self._value == cast(int, other)
